import errno
import logging
import threading

import av.error
import av.logging


def do_log(message: str) -> None:
    av.logging.log(av.logging.INFO, "test", message)


def test_adapt_level() -> None:
    assert av.logging.adapt_level(av.logging.ERROR) == logging.ERROR
    assert av.logging.adapt_level(av.logging.WARNING) == logging.WARNING
    assert (
        av.logging.adapt_level((av.logging.WARNING + av.logging.ERROR) // 2)
        == logging.WARNING
    )


def test_threaded_captures() -> None:
    av.logging.set_level(av.logging.VERBOSE)

    with av.logging.Capture(local=True) as logs:
        do_log("main")
        thread = threading.Thread(target=do_log, args=("thread",))
        thread.start()
        thread.join()

    assert (av.logging.INFO, "test", "main") in logs
    av.logging.set_level(None)


def test_global_captures() -> None:
    av.logging.set_level(av.logging.VERBOSE)

    with av.logging.Capture(local=False) as logs:
        do_log("main")
        thread = threading.Thread(target=do_log, args=("thread",))
        thread.start()
        thread.join()

    assert (av.logging.INFO, "test", "main") in logs
    assert (av.logging.INFO, "test", "thread") in logs
    av.logging.set_level(None)


def test_repeats() -> None:
    av.logging.set_level(av.logging.VERBOSE)

    with av.logging.Capture() as logs:
        do_log("foo")
        do_log("foo")
        do_log("bar")
        do_log("bar")
        do_log("bar")
        do_log("baz")

    logs = [log for log in logs if log[1] == "test"]

    assert logs == [
        (av.logging.INFO, "test", "foo"),
        (av.logging.INFO, "test", "foo"),
        (av.logging.INFO, "test", "bar"),
        (av.logging.INFO, "test", "bar (repeated 2 more times)"),
        (av.logging.INFO, "test", "baz"),
    ]

    av.logging.set_level(None)


def test_error() -> None:
    av.logging.set_level(av.logging.VERBOSE)

    log = (av.logging.ERROR, "test", "This is a test.")
    av.logging.log(*log)
    try:
        av.error.err_check(-errno.EPERM)
    except av.error.PermissionError as e:
        assert e.log == log
    else:
        assert False

    av.logging.set_level(None)



from fractions import Fraction

import pytest

import av
from av import AudioFrame, AudioResampler


def test_flush_immediately() -> None:
    """
    If we flush the resampler before passing any input, it returns
    a `None` frame without setting up the graph.
    """

    resampler = AudioResampler()

    # flush
    oframes = resampler.resample(None)
    assert len(oframes) == 0


def test_identity_passthrough() -> None:
    """
    If we don't ask it to do anything, it won't.
    """

    resampler = AudioResampler()

    # resample one frame
    iframe = AudioFrame("s16", "stereo", 1024)

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1
    assert iframe is oframes[0]

    # resample another frame
    iframe.pts = 1024

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1
    assert iframe is oframes[0]

    # flush
    oframes = resampler.resample(None)
    assert len(oframes) == 0


def test_matching_passthrough() -> None:
    """
    If the frames match, it won't do anything.
    """

    resampler = AudioResampler("s16", "stereo")

    # resample one frame
    iframe = AudioFrame("s16", "stereo", 1024)

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1
    assert iframe is oframes[0]

    # resample another frame
    iframe.pts = 1024

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1
    assert iframe is oframes[0]

    # flush
    oframes = resampler.resample(None)
    assert len(oframes) == 0


def test_pts_assertion_same_rate() -> None:
    av.logging.set_level(av.logging.VERBOSE)
    resampler = AudioResampler("s16", "mono")

    # resample one frame
    iframe = AudioFrame("s16", "stereo", 1024)
    iframe.sample_rate = 48000
    iframe.time_base = Fraction(1, 48000)
    iframe.pts = 0

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 0
    assert oframe.time_base == iframe.time_base
    assert oframe.sample_rate == iframe.sample_rate
    assert oframe.samples == iframe.samples

    # resample another frame
    iframe.pts = 1024

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 1024
    assert oframe.time_base == iframe.time_base
    assert oframe.sample_rate == iframe.sample_rate
    assert oframe.samples == iframe.samples

    # resample another frame with a pts gap, do not raise exception
    iframe.pts = 9999
    oframes = resampler.resample(iframe)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 9999
    assert oframe.time_base == iframe.time_base
    assert oframe.sample_rate == iframe.sample_rate
    assert oframe.samples == iframe.samples

    # flush
    oframes = resampler.resample(None)
    assert len(oframes) == 0
    av.logging.set_level(None)


def test_pts_assertion_new_rate_up() -> None:
    resampler = AudioResampler("s16", "mono", 44100)

    # resample one frame
    iframe = AudioFrame("s16", "stereo", 1024)
    iframe.sample_rate = 48000
    iframe.time_base = Fraction(1, 48000)
    iframe.pts = 0

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 0
    assert oframe.time_base == Fraction(1, 44100)
    assert oframe.sample_rate == 44100
    assert oframe.samples == 925

    iframe = AudioFrame("s16", "stereo", 1024)
    iframe.sample_rate = 48000
    iframe.time_base = Fraction(1, 48000)
    iframe.pts = 1024

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 925
    assert oframe.time_base == Fraction(1, 44100)
    assert oframe.sample_rate == 44100
    assert oframe.samples == 941

    # flush
    oframes = resampler.resample(None)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 941 + 925
    assert oframe.time_base == Fraction(1, 44100)
    assert oframe.sample_rate == 44100
    assert oframe.samples == 15


def test_pts_assertion_new_rate_down() -> None:
    resampler = AudioResampler("s16", "mono", 48000)

    # resample one frame
    iframe = AudioFrame("s16", "stereo", 1024)
    iframe.sample_rate = 44100
    iframe.time_base = Fraction(1, 44100)
    iframe.pts = 0

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 0
    assert oframe.time_base == Fraction(1, 48000)
    assert oframe.sample_rate == 48000
    assert oframe.samples == 1098

    iframe = AudioFrame("s16", "stereo", 1024)
    iframe.sample_rate = 44100
    iframe.time_base = Fraction(1, 44100)
    iframe.pts = 1024

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 1098
    assert oframe.time_base == Fraction(1, 48000)
    assert oframe.sample_rate == 48000
    assert oframe.samples == 1114

    # flush
    oframes = resampler.resample(None)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 1114 + 1098
    assert oframe.time_base == Fraction(1, 48000)
    assert oframe.sample_rate == 48000
    assert oframe.samples == 18


def test_pts_assertion_new_rate_fltp() -> None:
    resampler = AudioResampler("fltp", "mono", 8000, 1024)

    # resample one frame
    iframe = AudioFrame("s16", "mono", 1024)
    iframe.sample_rate = 8000
    iframe.time_base = Fraction(1, 1000)
    iframe.pts = 0

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 0
    assert oframe.time_base == Fraction(1, 8000)
    assert oframe.sample_rate == 8000
    assert oframe.samples == 1024

    iframe = AudioFrame("s16", "mono", 1024)
    iframe.sample_rate = 8000
    iframe.time_base = Fraction(1, 1000)
    iframe.pts = 8192

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 65536
    assert oframe.time_base == Fraction(1, 8000)
    assert oframe.sample_rate == 8000
    assert oframe.samples == 1024

    # flush
    oframes = resampler.resample(None)
    assert len(oframes) == 0


def test_pts_missing_time_base() -> None:
    resampler = AudioResampler("s16", "mono", 44100)

    # resample one frame
    iframe = AudioFrame("s16", "stereo", 1024)
    iframe.sample_rate = 48000
    iframe.pts = 0

    oframes = resampler.resample(iframe)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 0
    assert oframe.time_base == Fraction(1, 44100)
    assert oframe.sample_rate == 44100

    # flush
    oframes = resampler.resample(None)
    assert len(oframes) == 1

    oframe = oframes[0]
    assert oframe.pts == 925
    assert oframe.time_base == Fraction(1, 44100)
    assert oframe.sample_rate == 44100
    assert oframe.samples == 16


def test_mismatched_input() -> None:
    """
    Consecutive frames must have the same layout, sample format and sample rate.
    """
    resampler = AudioResampler("s16", "mono", 44100)

    # resample one frame
    iframe = AudioFrame("s16", "stereo", 1024)
    iframe.sample_rate = 48000
    resampler.resample(iframe)

    # resample another frame with a sample format
    iframe = AudioFrame("s16", "mono", 1024)
    iframe.sample_rate = 48000
    with pytest.raises(
        ValueError, match="Frame does not match AudioResampler setup."
    ) as cm:
        resampler.resample(iframe)



import sys

import pytest

import av.error
from av.device import DeviceInfo, enumerate_input_devices, enumerate_output_devices


def test_device_info_attributes() -> None:
    d = DeviceInfo("0", "FaceTime HD Camera", True, ["video"])
    assert d.name == "0"
    assert d.description == "FaceTime HD Camera"
    assert d.is_default is True
    assert d.media_types == ["video"]


def test_device_info_repr_default() -> None:
    d = DeviceInfo("0", "FaceTime HD Camera", True, ["video"])
    assert repr(d) == "<av.DeviceInfo '0' 'FaceTime HD Camera' (default)>"


def test_device_info_repr_non_default() -> None:
    d = DeviceInfo("1", "Built-in Microphone", False, ["audio"])
    assert repr(d) == "<av.DeviceInfo '1' 'Built-in Microphone'>"


def test_enumerate_input_devices_unknown_format() -> None:
    with pytest.raises(ValueError, match="no such input format"):
        enumerate_input_devices("not_a_real_format_xyz")


def test_enumerate_output_devices_unknown_format() -> None:
    with pytest.raises(ValueError, match="no such output format"):
        enumerate_output_devices("not_a_real_format_xyz")


def _assert_valid_device_list(devices: list[DeviceInfo]) -> None:
    assert isinstance(devices, list)
    for device in devices:
        assert isinstance(device, DeviceInfo)
        assert isinstance(device.name, str)
        assert isinstance(device.description, str)
        assert isinstance(device.is_default, bool)
        assert isinstance(device.media_types, list)
        assert all(isinstance(mt, str) for mt in device.media_types)


@pytest.mark.skipif(sys.platform != "darwin", reason="avfoundation is macOS only")
def test_enumerate_input_devices_avfoundation() -> None:
    _assert_valid_device_list(enumerate_input_devices("avfoundation"))


@pytest.mark.skipif(sys.platform != "linux", reason="v4l2 is Linux only")
def test_enumerate_input_devices_v4l2() -> None:
    try:
        _assert_valid_device_list(enumerate_input_devices("video4linux2"))
    except av.error.OSError:
        pytest.skip("v4l2 device enumeration not available")


@pytest.mark.skipif(sys.platform != "win32", reason="dshow is Windows only")
def test_enumerate_input_devices_dshow() -> None:
    try:
        _assert_valid_device_list(enumerate_input_devices("dshow"))
    except av.error.OSError:
        pytest.skip("dshow device enumeration not available")
